==> jpaservices-11.1-sources.jar: org/fryske_akademy/jpa/JpqlBuilderImpl.java <==
/*
 * Copyright 2018 Fryske Akademy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.fryske_akademy.jpa;

/*-
 * #%L
 * jpaservices
 * %%
 * Copyright (C) 2018 Fryske Akademy
 * %%
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * #L%
 */
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;
import javax.enterprise.context.ApplicationScoped;
import javax.inject.Named;
import javax.persistence.Parameter;
import javax.persistence.Query;
import org.fryske_akademy.services.CrudReadService.SORTORDER;

/**
 * Stateless, threadsafe builder containing functions for building jpql}. 
 * Supports all comparison
 * expressions except "between" when building jpql.
 *
 * @author eduard
 */
@ApplicationScoped
@Named
public class JpqlBuilderImpl implements JpqlBuilder {

    private static final Logger LOGGER = Logger.getLogger(JpqlBuilderImpl.class.getName());

    /**
     *
     * builds an order by clause, uses {@link #ENTITY_PREFIX}.
     *
     * @param sort
     * @return
     */
    @Override
    public String orderClause(Map<String, SORTORDER> sort) {
        if (sort != null && !sort.isEmpty()) {
            StringBuilder srt = new StringBuilder("");
            boolean first = true;
            for (Map.Entry<String, SORTORDER> entry : sort.entrySet()) {
                String sortField = entry.getKey();
                if (sortField == null || entry.getValue() == SORTORDER.NONE) {
                    continue;
                }
                srt.append((first) ? " order by" : ",")
                        .append(ENTITY_PREFIX).append(sortField).append(" ").append(entry.getValue());
                first = false;
            }
            return srt.toString();
        }
        return "";
    }

    /**
     * builds a where clause, calls {@link #whereCondition(org.fryske_akademy.services.Param)
     * }
     * for every entry, uses {@link #TABLE_ALIAS} as alias for the table
     *
     * @param params
     * @return
     */
    @Override
    public String whereClause(List<Param> params) {
        if (params != null && !params.isEmpty()) {
            StringBuilder where = new StringBuilder("");
            boolean first = true;
            for (Param param : params) {
                where.append((first) ? " where" : param.getAndOr()).append(whereCondition(param));
                first = false;
            }
            return where.toString();
        }
        return "";
    }

    /**
     * builds a where condition, prepares query parameters later filled in
     * {@link #setWhereParams(javax.persistence.Query, java.util.List) }. "member of" support for singular param value and for a collection
     * as param value, when param value is a collection multiple "member of" conditions are generated.
     * "between" is not supported.
     *
     * @param param
     * @return
     */
    @Override
    public String whereCondition(Param param) {
        if (param.getOperator().toLowerCase().contains("member of")) {
            if (param.getParamValue() instanceof Collection) {
                if (!((Collection) param.getParamValue()).isEmpty()) {
                    StringBuilder rv = new StringBuilder(" (");
                    AtomicBoolean first = new AtomicBoolean(true);
                    AtomicInteger i = new AtomicInteger();
                    ((Collection) param.getParamValue()).forEach((t) -> {
                        if (first.get()) {
                            first.set(false);
                        } else {
                            rv.append(" or ");
                        }
                        rv.append(":").append(param.getParamKey()).append(i.getAndIncrement()).append(param.getNot()).append(param.getOperator()).append(ENTITY_PREFIX).append(param.getPropertyPath());
                    });
                    return rv.append(')').toString();
                } else {
                    return "";
                }
            } else {
                return " :" + param.getParamKey() + param.getNot() + param.getOperator() + ENTITY_PREFIX + param.getPropertyPath();
            }
        } else if (param.getOperator().toLowerCase().contains("in")) {
            return ENTITY_PREFIX + param.getPropertyPath() + param.getNot() + param.getOperator() + "( :" + param.getParamKey() + " )";
        } else if (param.getOperator().toLowerCase().contains("null") || param.getOperator().toLowerCase().contains("empty")) {
            return ENTITY_PREFIX + param.getPropertyPath() + param.getOperator();
        } else {
            return param.isCaseInsensitive() && String.class.equals(param.getParamType())
                    ? param.getNot() + "lower(" + ENTITY_PREFIX + param.getPropertyPath() + ")" + param.getOperator() + "lower(:" + param.getParamKey() + ")"
                    : param.getNot() + ENTITY_PREFIX + param.getPropertyPath() + param.getOperator() + ":" + param.getParamKey();
        }
    }
    
    public static final String ENTITY_PREFIX = " e.";

    /**
     * Calls {@link #setParam(javax.persistence.Query, org.fryske_akademy.services.Param)
     * } for each filter.
     *
     * @param q
     * @param params
     */
    @Override
    public void setWhereParams(Query q, List<Param> params) {
        if (params != null && !params.isEmpty()) {
            params.forEach((param) -> {
                setParam(q, param);
            });
        }
    }

    /**
     * When the Query has the parameter "key" call {@link Query#setParameter(java.lang.String, java.lang.Object)
     * }
     * otherwise call {@link Query#setParameter(int, java.lang.Object) } with
     * Short.valueOf(key), because then a positional parameter in a native query
     * is assumed.
     *
     * @param q
     * @param key
     * @param value
     */
    protected void set(Query q, String key, Object value) {
        if (isNamedParameter(q, key)) {
            q.setParameter(key, value);
        } else {
            if (LOGGER.isLoggable(Level.FINE)) {
                LOGGER.fine(String.format("%s not found in query, assuming positional parameter with value %s", key, value));
            }
            q.setParameter(Short.valueOf(key), value);
        }
    }

    private static final Pattern NUM = Pattern.compile("[1-9]+");

    private static boolean containsKey(Query q, String key) {
        boolean num = NUM.matcher(key).matches();
        Integer i = num ? Integer.valueOf(key) : 0;
        return q.getParameters().stream().anyMatch((p) -> {
            return key.equals(p.getName()) || (num && i.equals(p.getPosition()));
        });
    }

    private static boolean isNamedParameter(Query q, String key) {
        return q.getParameters().stream().anyMatch((p) -> {
            return key.equals(p.getName());
        });
    }

    /**
     * Fills parameters prepared in {@link #setWhereParams(javax.persistence.Query, java.util.List)
     * }, if the type of the field in the query is a Short or an Integer and the
     * paramValue is a String, it is converted accordingly. For native queries {@link Param#getParamKey()
     * } should be a numeric positional parameter. When the operator is member
     * of and the value is a Collection a value is set for each entry in the
     * collection, for this a parameter key {@link Param#getParamKey() }
     * with an index number appended is assumed, {@link #whereCondition(org.fryske_akademy.services.Param)
     * } prepares this. For the type check either
     * {@link Parameter#getParameterType() } is used, or, when null, {@link Param#getParamType()
     * }. Via
     * Param you have full control over the Object that is used as paramValue.
     *
     * @param q
     * @param param
     */
    @Override
    public void setParam(Query q, Param param) {
        if (!containsKey(q, param.getParamKey()) && !(param.getOperator().toLowerCase().contains("member of") && param.getParamValue() instanceof Collection)) {
            if (LOGGER.isLoggable(Level.FINE)) {
                LOGGER.fine(String.format("parameter not found in query, skipping %s", param));
            }
            return;
        }
        if (LOGGER.isLoggable(Level.FINE)) {
            LOGGER.fine("trying to set: " + param);
        }
        if (param.getOperator().toLowerCase().contains("member of") && param.getParamValue() instanceof Collection) {
            AtomicInteger i = new AtomicInteger();
            ((Collection) param.getParamValue()).forEach((t) -> {
                set(q, param.getParamKey() + (i.getAndIncrement()), t);
            });
        } else {
            if (param.getParamValue() instanceof Collection && !param.getOperator().toLowerCase().contains("in")) {
                if (LOGGER.isLoggable(Level.WARNING)) {
                    LOGGER.warning("value for " + param.getParamKey() + " is a collection, you may want to use the \"in\" operator");
                }
            }
            set(q, param.getParamKey(), param.getParamValue());
        }
    }

}

==> jpaservices-12.1-sources.jar: org/fryske_akademy/jpa/JpqlBuilderImpl.java <==
/*
 * Copyright 2018 Fryske Akademy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.fryske_akademy.jpa;

/*-
 * #%L
 * jpaservices
 * %%
 * Copyright (C) 2018 Fryske Akademy
 * %%
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * #L%
 */

import org.fryske_akademy.services.CrudReadService.SORTORDER;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.enterprise.context.ApplicationScoped;
import javax.inject.Named;
import javax.persistence.Parameter;
import javax.persistence.Query;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * Stateless, threadsafe builder containing functions for building jpql}. 
 * Supports all comparison
 * expressions except "between" when building jpql.
 *
 * @author eduard
 */
@ApplicationScoped
@Named
public class JpqlBuilderImpl implements JpqlBuilder {

    private static final Logger LOGGER = LoggerFactory.getLogger(JpqlBuilderImpl.class.getName());
    public static final String ENTITY_PREFIX = " e.";
    private static boolean isNum(String s) {
        try {
            Integer.parseInt(s);
        } catch (NumberFormatException ex) {
            return false;
        }
        return true;
    }

    private static boolean containsKey(Query q, String key) {
        boolean num = isNum(key);
        Integer i = num ? Integer.valueOf(key) : 0;
        return q.getParameters().stream().anyMatch((p) -> {
            return key.equals(p.getName()) || (num && i.equals(p.getPosition()));
        });
    }
    private static boolean isNamedParameter(Query q, String key) {
        return q.getParameters().stream().anyMatch((p) -> {
            return key.equals(p.getName());
        });
    }

    /**
     *
     * builds an order by clause, uses {@link #ENTITY_PREFIX}.
     *
     * @param sort
     * @return
     */
    @Override
    public String orderClause(Map<String, SORTORDER> sort) {
        if (sort != null && !sort.isEmpty()) {
            StringBuilder srt = new StringBuilder();
            boolean first = true;
            for (Map.Entry<String, SORTORDER> entry : sort.entrySet()) {
                String sortField = entry.getKey();
                if (sortField == null || entry.getValue() == SORTORDER.NONE) {
                    continue;
                }
                srt.append((first) ? " order by" : ",")
                        .append(ENTITY_PREFIX).append(sortField).append(" ").append(entry.getValue());
                first = false;
            }
            return srt.toString();
        }
        return "";
    }

    /**
     * builds a where clause, calls {@link #whereCondition(Param)}
     * for every entry.
     *
     * @param params
     * @return
     */
    @Override
    public String whereClause(List<Param> params) {
        if (params != null && !params.isEmpty()) {
            StringBuilder where = new StringBuilder();
            boolean first = true;
            for (Param param : params) {
                where.append((first) ? " where" : param.getAndOr()).append(whereCondition(param));
                first = false;
            }
            return where.toString();
        }
        return "";
    }
    
    

    /**
     * builds a where condition, prepares query parameters later filled in
     * {@link #setWhereParams(javax.persistence.Query, java.util.List) }. "member of" support for singular param value and for a collection
     * as param value, when param value is a collection multiple "member of" conditions are generated.
     * "between" is not supported.
     *
     * @param param
     * @return
     */
    @Override
    public String whereCondition(Param param) {
        if (param.getOperator().toLowerCase().contains("member of")) {
            if (param.getParamValue() instanceof Collection) {
                if (!((Collection) param.getParamValue()).isEmpty()) {
                    StringBuilder rv = new StringBuilder(" (");
                    AtomicBoolean first = new AtomicBoolean(true);
                    AtomicInteger i = new AtomicInteger();
                    ((Collection) param.getParamValue()).forEach((t) -> {
                        if (first.get()) {
                            first.set(false);
                        } else {
                            rv.append(" or ");
                        }
                        rv.append(":").append(param.getParamKey()).append(i.getAndIncrement()).append(param.getNot()).append(param.getOperator()).append(ENTITY_PREFIX).append(param.getPropertyPath());
                    });
                    return rv.append(')').toString();
                } else {
                    return "";
                }
            } else {
                return " :" + param.getParamKey() + param.getNot() + param.getOperator() + ENTITY_PREFIX + param.getPropertyPath();
            }
        } else if (param.getOperator().toLowerCase().contains("in")) {
            return ENTITY_PREFIX + param.getPropertyPath() + param.getNot() + param.getOperator() + "( :" + param.getParamKey() + " )";
        } else if (param.getOperator().toLowerCase().contains("null") || param.getOperator().toLowerCase().contains("empty")) {
            return ENTITY_PREFIX + param.getPropertyPath() + param.getOperator();
        } else {
            // here we may
            return param.isCaseInsensitive() && String.class.equals(param.getParamType())
                    ? param.getNot() + "lower(" + ENTITY_PREFIX + param.getPropertyPath() + ")" + param.getOperator() + "lower(:" + param.getParamKey() + ")"
                    : param.getNot() + ENTITY_PREFIX + param.getPropertyPath() + param.getOperator() + ":" + param.getParamKey();
        }
    }
    

    /**
     * Calls {@link #setParam(Query, Param)}  for each filter.
     *
     * @param q
     * @param params
     */
    @Override
    public void setWhereParams(Query q, List<Param> params) {
        if (params != null && !params.isEmpty()) {
            params.forEach((param) -> {
                setParam(q, param);
            });
        }
    }

    /**
     * When the Query has the parameter "key" call {@link Query#setParameter(java.lang.String, java.lang.Object)
     * }
     * otherwise call {@link Query#setParameter(int, java.lang.Object) } with
     * Short.valueOf(key), because then a positional parameter in a native query
     * is assumed.
     *
     * @param q
     * @param key
     * @param value
     */
    protected void set(Query q, String key, Object value) {
        if (isNamedParameter(q, key)) {
            q.setParameter(key, value);
        } else {
            if (LOGGER.isDebugEnabled()) {
                LOGGER.debug(String.format("%s not found in query, assuming positional parameter with value %s", key, value));
            }
            q.setParameter(Short.valueOf(key), value);
        }
    }


    /**
     * Fills parameters prepared in {@link #setWhereParams(javax.persistence.Query, java.util.List)
     * }, if the type of the field in the query is a Short or an Integer and the
     * paramValue is a String, it is converted accordingly. For native queries {@link Param#getParamKey()
     * } should be a numeric positional parameter. When the operator is member
     * of and the value is a Collection a value is set for each entry in the
     * collection, for this a parameter key {@link Param#getParamKey() }
     * with an index number appended is assumed, {@link #whereCondition(Param)}  prepares this. For the type check either
     * {@link Parameter#getParameterType() } is used, or, when null, {@link Param#getParamType()
     * }. Via
     * Param you have full control over the Object that is used as paramValue.
     *
     * @param q
     * @param param
     */
    @Override
    public void setParam(Query q, Param param) {
        if (!containsKey(q, param.getParamKey()) && !(param.getOperator().toLowerCase().contains("member of") && param.getParamValue() instanceof Collection)) {
            if (LOGGER.isDebugEnabled()) {
                LOGGER.debug(String.format("parameter not found in query, skipping %s", param));
            }
            return;
        }
        if (LOGGER.isDebugEnabled()) {
            LOGGER.debug("trying to set: " + param);
        }
        if (param.getOperator().toLowerCase().contains("member of") && param.getParamValue() instanceof Collection) {
            AtomicInteger i = new AtomicInteger();
            ((Collection) param.getParamValue()).forEach((t) -> {
                set(q, param.getParamKey() + (i.getAndIncrement()), t);
            });
        } else {
            if (param.getParamValue() instanceof Collection && !param.getOperator().toLowerCase().contains("in")) {
                if (LOGGER.isWarnEnabled()) {
                    LOGGER.warn("value for " + param.getParamKey() + " is a collection, you may want to use the \"in\" operator");
                }
            }
            set(q, param.getParamKey(), param.getParamValue());
        }
    }

}
